/*
 * QEMU4V tracer
 *
 *  Copyright (c) 2015-2018 Sergey Smolov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */

#include "include/q4v_tracer.h"

#include "qemu/osdep.h"
#include "qemu-common.h"
#include "qemu/log.h"

/* Flag that enables QEMU4V tracing. */
static bool trace_en = false;

void enable_q4v_trace(void) {
    trace_en = true;
}

bool q4v_trace_enabled(void) {
    return trace_en;
}

/* Flag that enables Page Table Entry addresses printing (for Aarch64 only). */
static bool print_pte_addr_en = false;

void enable_q4v_pte(void) {
    print_pte_addr_en = true;
}

bool q4v_pte_enabled(void) {
    return print_pte_addr_en;
}

static int timestamp;

int q4v_tstamp(void) {
    return timestamp;
}

void set_q4v_tstamp(int tstamp) {
    timestamp = tstamp;
}

static char *reg_write_prefix;

/* Flag that shows that write to some register is logged right now. */
static int trace_reg_logged;

/*
 * Internal function.
 */
void q4v_trace_write_regname_prefix(const char *regname) {

    if (q4v_trace_enabled()) {

        /* Start register record. */
        trace_reg_logged = 1;
        reg_write_prefix = (char *)malloc(256*sizeof(char));
        sprintf(reg_write_prefix, "%d clk R %s", q4v_tstamp(), regname);
    }
}

void log_write_prefix(const int reg) {
    char regname[4];
    sprintf(regname, "x%d", reg);

    q4v_trace_write_regname_prefix(regname);
}

/*
 * Internal function.
 */
void q4v_trace_value(const uint64_t value) {

    if (q4v_trace_enabled() && trace_reg_logged) {
        qemu_log_lock();
        qemu_log("%s %016"PRIx64"\n", reg_write_prefix, value);
        qemu_log_unlock();
        trace_reg_logged = 0;
        free(reg_write_prefix);
    }
}

void q4v_trace_reg_write(const char *regname, const uint64_t value) {
    q4v_trace_write_regname_prefix(regname);
    q4v_trace_value(value);
}

void q4v_trace_mem_access(const char *rw,
    const int size,
    const uint64_t mem_addr,
    const uint64_t data) {

    if (q4v_trace_enabled()) {
        qemu_log_lock();
        qemu_log("%d clk M%s%d %.16"PRIx64" %.16"PRIx64"\n",
            q4v_tstamp(), rw, size, mem_addr, data);
        qemu_log_unlock();
    }
}

char *trimwhitespace(char *str)
{
    char *end;

    // Trim leading space
    while(isspace((unsigned char)*str)) str++;

    if(*str == 0)  // All spaces?
        return str;

    // Trim trailing space
    end = str + strlen(str) - 1;
    while(end > str && isspace((unsigned char)*end)) end--;

    // Write new null terminator
    *(end+1) = 0;

    return str;
}

char* insn_strcat(char *insn_str, const char *str) {

    int insn_strlen = strlen(insn_str);
    int str_strlen = strlen(str);

    if (insn_strlen + str_strlen > 256) {
        qemu_log("%s: string overflow at: insn_str=%s, str=%s\n", __func__, insn_str, str);
    }

    return strcat(insn_str, str);
}
